import yaml

from remotemanager.serialisation.serial import serial


class serialyaml(serial):
    """
    subclass of serial, implementing yaml methods
    """

    def dumps(self, obj):
        obj = self.wrap_to_list(obj)
        return yaml.dump(obj)

    def loads(self, string):
        loaded = yaml.safe_load(string)

        return self.wrap_to_list(loaded)

    @property
    def callstring(self):
        return "yaml"

    @property
    def loadstring(self) -> str:
        return "safe_load"

    def dumpfunc(self) -> str:
        lines = [
            "\ndef dump(obj, file):",
            f"\t{self.importstring}",
            f"\tif isinstance(obj, (set, tuple)):",
            f"\t\tobj = list(obj)",
            f'\tif not file.endswith("{self.extension}"):',
            f'\t\tfile = file + "{self.extension}"',
            f'\twith open(file, "{self.write_mode}") as o:',
            f"\t\t{self.callstring}.{self.dumpstring}(obj, o)",
        ]

        return "\n".join(lines)



import importlib
import inspect
import os
import typing

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils.uuid import generate_uuid
from remotemanager.logging import LoggingMixin


_SCRIPT_TEMPLATE = """
{function}

if __name__ == '__main__':
\tkwargs = {args}

\tresult = {name}(**kwargs)
"""


class Function(SendableMixin, LoggingMixin):
    """
    Serialise function to an executable python file for transfer

    Args:
        func:
            python function for serialisation
    """

    def __init__(self, func: [typing.Callable, str]):
        self._logger.debug(f"creating new serialisable function for {func}")

        self._uuid = ""
        try:
            source = inspect.getsource(func)
            source = source.replace("@RemoteFunction", "").strip()

            self._signature = Function.prepare_signature(inspect.signature(func))

            self._logger.debug(f"updated signature to {self._signature}")
            rawsig = Function.get_raw_signature(source)
            self._logger.debug(f"raw typed signature detected as" f" {self._signature}")
            source = source.replace(rawsig, self._signature, 1)

            self._source = source
            self._fname = func.__name__

        except TypeError as e:
            if isinstance(func, str):
                self._source = func

                # need to extract the name
                name = "f"
                signature = ")"
                for line in func.split("\n"):
                    if line.startswith("def"):
                        name, signature = line.split("def ")[-1].split("(")
                self._fname = name

                self._signature = "(" + signature.strip(":")
            else:
                raise e
        self._uuid = generate_uuid(self._source)

    def __call__(self, *args, **kwargs):
        return self.object(*args, **kwargs)

    @staticmethod
    def get_raw_signature(source):
        """
        Strips the signature as it is typed.
        inspect.signature does some formatting which can cause replacement to
        break in some conditions

        Args:
            source (str):
                raw source

        Returns:
            (str): raw signature as typed
        """
        definition = []
        for line in source.split("\n"):
            # we need to split from the function name for the raw signature
            if "(" in line:
                # split from first ( and up,
                # then join for any tuples defined as default
                line = "(" + "(".join(line.split("(")[1:])

            # appending the final ":" causes issues, strip that in a safe
            # manner then append
            definition.append(":".join(line.split(":")[:-1]))

            # we've reached the end of the definition
            if line.endswith(":"):
                break

        return "\n".join(definition)

    @staticmethod
    def prepare_signature(sig) -> str:
        """
        Inserts *args and **kwargs into any signature that does not already
        have it

        Args:
            sig:
                inspect.signature(func)

        Returns:
            (str): formatted sig
        """

        args = [str(a) for a in sig.parameters.values()]

        if "*args" not in args:
            if "**kwargs" in args:
                args.insert(-1, "*args")
            else:
                args.append("*args")

        if "**kwargs" not in args:
            args.append("**kwargs")

        return f'({", ".join(args)})'

    @property
    def name(self):
        """
        Function name
        """
        return self._fname

    @property
    def raw_source(self):
        """
        Function source
        """
        return self._source

    @property
    def source(self):
        """
        Function source

        Returns:
            (str): source code
        """
        return self._source

    @property
    def signature(self):
        return self._signature

    @property
    def uuid(self):
        """
        Function uuid (64 characters)
        """
        return self._uuid

    @property
    def object(self):
        """
        Recreates the function object by writing out the source, and importing.

        Returns:
            typing.Callable:
                the originally passed function
        """

        tmp_file = ""
        try:
            tmp_file = os.path.abspath(f"{self.uuid}.py")

            with open(tmp_file, "w+") as o:
                o.write(self.source)

            func_module = importlib.import_module(self.uuid)
            func_object = getattr(func_module, f"{self.name}")

        finally:
            os.remove(tmp_file)

        return func_object

    def dump_to_string(self, args):
        """
        Dump this function to a serialised string, ready to be written to a
        python file

        Args:
            args (dict):
                arguments to be used for this dumped run

        Returns:
            (str):
                serialised file
        """

        if args is None:
            args = {}

        return _SCRIPT_TEMPLATE.format(
            **{"function": self.source, "name": self.name, "args": args}
        )



"""
Baseclass for any file transfer
"""
import os.path

from remotemanager.storage.trackedfile import TrackedFile
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils import ensure_list, ensure_dir
from remotemanager.logging import LoggingMixin
from remotemanager.utils.flags import Flags
from remotemanager.connection import URL


class Transport(SendableMixin, LoggingMixin):
    """
    Baseclass for file transfer

    Args:
        url (URL):
            url to extract remote address from
    """

    def __init__(self, url=None, *args, **kwargs):
        self._remote_address = None
        self._url = url if url is not None else URL()
        self.set_remote(url)

        self._flags = Flags()
        self._transfers = {}

    def queue_for_push(
        self, files: [list, str, TrackedFile], local: str = None, remote: str = None
    ):
        """
        Queue file(s) for sending (pushing)

        Args:
            files (list[str], str, TrackedFile):
                list of files (or file) to add to push queue
            local (str):
                local/origin folder for the file(s)
            remote (str):
                remote/destination folder for the file(s)
        Returns:
            None
        """
        if isinstance(files, TrackedFile):
            self._logger.info(f"adding TrackedFile {files.name} to PUSH queue")
            self.add_transfer(files.name, files.local_dir, files.remote_dir, "push")
            return
        self._logger.info(f"adding to PUSH queue")
        self.add_transfer(files, local, remote, "push")

    def queue_for_pull(
        self, files: [list, str, TrackedFile], local: str = None, remote: str = None
    ):
        """
        Queue file(s) for retrieving (pulling)

        Args:
            files (list[str], str, TrackedFile):
                list of files (or file) to add to pull queue
            local (str):
                local/destination folder for the file(s)
            remote (str):
                remote/origin folder for the file(s)
        Returns:
            None
        """
        if isinstance(files, TrackedFile):
            self._logger.info(f"adding TrackedFile {files.name} to PULL queue")
            self.add_transfer(files.name, files.remote_dir, files.local_dir, "pull")
            return
        self._logger.info(f"adding to PULL queue")
        self.add_transfer(files, remote, local, "pull")

    def add_transfer(self, files: [list, str], origin: str, target: str, mode: str):
        """
        Create a transfer to be executed. The ordering of the origin/target
        files should be considered as this transport instance being a
        "tunnel" between wherever it is executed (origin), and the destination
        (target)

        Args:
            files (list[str], str):
                list of files (or file) to add to pull queue
            origin (str):
                origin folder for the file(s)
            target (str):
                target folder for the file(s)
            mode (str: "push" or "pull"):
                transfer mode. Chooses where the remote address is placed
        Returns:
            None
        """
        modes = ("push", "pull")
        if mode.lower() not in modes:
            raise ValueError(f"mode must be one of {modes}")

        if origin is None:
            origin = "."
        if target is None:
            target = "."

        # ensure dir-type
        origin = os.path.join(origin, "")
        target = os.path.join(target, "")

        if mode == "push":
            pair = f"{origin}>{self._add_address(target)}"
        else:
            pair = f"{self._add_address(origin)}>{target}"

        files = [os.path.split(f)[1] for f in ensure_list(files)]

        self._logger.info(
            f"adding transfer: {Transport.split_pair(pair)[0]} "
            f"-> {Transport.split_pair(pair)[1]}"
        )
        self._logger.info(f"for files {files}")

        if pair in self._transfers:
            self._transfers[pair] = self._transfers[pair].union(set(files))
        else:
            self._transfers[pair] = set(files)

    def _add_address(self, dir: str) -> str:
        """
        Adds the remote address to the dir `dir` if it exists

        Args:
            dir (str):
                remote dir to have address appended

        Returns:
            (str) dir
        """
        if self.address is None:
            return dir
        return f"{self.address}:{dir}"

    @staticmethod
    def _format_for_cmd(folder: str, inp: list) -> str:
        """
        Formats a list into a bash expandable command with brace expansion

        Args:
            folder (str):
                the dir to copy to/from
            inp (list):
                list of items to compress

        Returns (str):
            formatted cmd
        """

        if isinstance(inp, str):
            raise ValueError(
                "files is stringtype, " "was a transfer forced into the queue?"
            )

        if len(inp) > 1:
            return os.path.join(folder, "{" + ",".join(inp) + "}")
        return os.path.join(folder, inp[0])

    @property
    def transfers(self) -> dict:
        """
        Return the current transfer dict

        Returns (dict):
            {paths: files} transfer dict
        """
        return {k: sorted(list(v)) for k, v in self._transfers.items()}

    def print_transfers(self):
        """
        Print a formatted version of the current queued transfers

        Returns:
            None
        """
        i = 0
        for pair, files in self.transfers.items():
            i += 1
            print(
                f"transfer {i}:"
                f"\norigin: {Transport.split_pair(pair)[0]}"
                f"\ntarget: {Transport.split_pair(pair)[1]}"
            )
            j = 0
            for file in files:
                j += 1
                print(f"\t({j}/{len(files)}) {file}")

    @property
    def address(self):
        """
        return the remote address

        Returns (str):
            the remote address
        """
        return self._remote_address

    @address.setter
    def address(self, remote_address):
        """
        set the remote address

        Returns:
            None
        """
        self._remote_address = remote_address

    @property
    def url(self):
        if self._url is not None:
            return self._url

        return URL()

    @url.setter
    def url(self, url):
        self._url = url

    def set_remote(self, url=None):
        """
        set the remote address with a URL object

        Returns:
            None
        """
        self._logger.info(f"setting rsync url to {url}")
        if url is None:
            self._logger.info(f"url is None, setting None")
            self._remote_address = None
        elif url.is_local:
            self._logger.info(f"url is local, setting None")
            self._remote_address = None
        else:
            self._logger.info(f"url okay, setting")
            self._remote_address = url.userhost
            self.url = url

    @property
    def flags(self):
        return self._flags

    @flags.setter
    def flags(self, new):
        self._flags = Flags(str(new))

    def cmd(self, primary, secondary):
        """
        Returns a formatted command for issuing transfers. It is left to
        the developer to implement this method when adding more transport
        classes.

        The implementation should take two strings as arguments, `primary` and
        `secondary`:

        `primary`: This is the source folder, containing the files for transfer

        `secondary`: This is the destination folder for the files

        At its most basic:

        ```
        def cmd(self, primary, secondary):
            base = self._cmd.format(primary=primary,
                                    secondary=secondary)
            return base
        ```

        where self._cmd is a string of the format

        `command {primary} {secondary}`

        You can, of course, extend upon this. View the included transport
        methods for ideas on how to do this.

        Returns (str):
            formatted command for issuing a transfer
        """
        raise NotImplementedError

    def transfer(
        self, dry_run: bool = False, prepend: bool = True, raise_errors: bool = True
    ):
        """
        Perform the actual transfer

        Args:
            dry_run (bool):
                do not perform command, just return the command(s) to be
                executed
            prepend (bool):
                enable forced cmd prepending
            raise_errors (bool):
                will not raise any stderr if False

        Returns (str, None):
            the dry run string, or None
        """

        self._logger.info(f"executing a transfer")

        commands = []
        for pair, files in self.transfers.items():
            primary, secondary = Transport.split_pair(pair)

            primary = self._format_for_cmd(primary, files)

            base_cmd = self.cmd(primary=primary, secondary=secondary)

            commands.append(base_cmd)

        if dry_run:
            return [
                self.url.cmd(cmd, dry_run=True, local=True, prepend=prepend)
                for cmd in commands
            ]

        for cmd in commands:
            self.url.cmd(cmd, local=True, prepend=prepend, raise_errors=raise_errors)
            # wipe the transfer queue
            self.wipe_transfers()

    def wipe_transfers(self):
        self._logger.info("wiping transfers")
        self._transfers = {}

    @staticmethod
    def split_pair(pair: str) -> list:
        """
        Convert a "dir>dir" string into list format

        Args:
            pair (tuple):
                (dir, dir) tuple to be split

        Returns (list):
            [dir, dir]

        """
        return [ensure_dir(os.path.split(p)[0]) for p in pair.split(">")]



"""
Handles file transfer via the `scp` protocol
"""
import os

from remotemanager.transport.transport import Transport
from remotemanager.utils import ensure_list


class scp(Transport):
    """
    Class to handle file transfers using the scp protocol

    Args:
        url (URL):
            url to extract remote address from
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # flags can be exposed, to utilise their flexibility
        flags = kwargs.pop("flags", "r")
        self.flags = flags

        self._transfers = {}

        self._cmd = "{create}{password}scp {sshkey}{flags} {primary} {secondary}"

    @staticmethod
    def _format_for_cmd(folder: str, inp: list) -> str:
        """
        Formats a list into a bash expandable command with brace expansion

        Args:
            folder (str):
                the dir to copy to/from
            inp (list):
                list of items to compress

        Returns (str):
            formatted cmd
        """

        def scp_join(files):
            return f" ".join([os.path.join(folder, f) for f in files])

        if isinstance(inp, str):
            raise ValueError(
                "files is stringtype, " "was a transfer forced into the queue?"
            )

        inp = ensure_list(inp)

        if ":" not in folder:
            return scp_join(inp)
        remote, folder = folder.split(":")
        return f'{remote}:"{scp_join(inp)}"'

    def cmd(self, primary, secondary):
        password = ""
        if self.url.passfile is not None:
            password = f"sshpass -f {self.url.passfile} "

        sshkey = ""
        if self.url.keyfile is not None:
            sshkey = f"-i {self.url.keyfile} "

        def get_remote_dir(path):
            if ":" not in path:
                return path
            return path.split(":")[1]

        mkdir = self.url.utils.mkdir(get_remote_dir(secondary), dry_run=True) + " && "

        base = self._cmd.format(
            create=mkdir,
            password=password,
            flags=self.flags,
            sshkey=sshkey,
            primary=primary,
            secondary=secondary,
        )
        self._logger.debug(f'returning formatted cmd: "{base}"')
        return base



from remotemanager.connection.computers.base import BaseComputer, required, optional
from remotemanager.connection.computers.parsers import torque, slurm


class Example_Torque(BaseComputer):
    """
    example class for connecting to a remote computer using a torque scheduler
    """

    def __init__(self, **kwargs):
        if "host" not in kwargs:
            kwargs["host"] = "remote.address.for.connection"

        super().__init__(**kwargs)

        self.submitter = "qsub"
        self.shebang = "#!/bin/bash"
        self._parser = torque

        self.mpi = required("ppn")
        self.omp = required("omp")
        self.nodes = required("nodes")
        self.queue = required("q")
        self.time = required("walltime")
        self.jobname = optional("N")
        self.outfile = optional("o")
        self.errfile = optional("e")

        self._internal_extra = ""


class Example_Slurm(BaseComputer):
    """
    example class for connecting to a remote computer using a slurm scheduler
    """

    def __init__(self, **kwargs):
        if "host" not in kwargs:
            kwargs["host"] = "remote.address.for.connection"

        super().__init__(**kwargs)

        self.submitter = "sbatch"
        self.shebang = "#!/bin/bash"
        self._parser = slurm

        self.mpi = required("ntasks")
        self.omp = required("cpus-per-task")
        self.nodes = required("nodes")
        self.queue = required("queue")
        self.time = required("walltime")
        self.jobname = optional("job-name")
        self.outfile = optional("output")
        self.errfile = optional("error")
